#include <windows.h>
#include <string.h>
#include "hem.h"
#include "strarray.h"

#define VERINFO_VERSION_MAJOR	1
#define VERINFO_VERSION_MINOR	0

#define	MAX_VERSIONINFOS		128	/* max number of VS_VERSIONINFO structures we can handle in a single module */

#define PAGES_BETWEEN_BREAK		100	/* number of pages/bytes to process between checks for user abort */
#define BYTES_BETWEEN_BREAK		0x1000 * PAGES_BETWEEN_BREAK

#define PHEM_BYTE HEM_BYTE *

/* global variables */
DWORD ViArray[MAX_VERSIONINFOS];	/* array of offsets of all available VS_VERSIONINFO structures */
int TotalVIs = 0;					/* total number of found VS_VERSIONINFO structures */
HANDLE hFile = 0;
HANDLE hMapping = 0;
LPBYTE lpMapping = 0;
DWORD FileLength;

/* a small preprocessor magic that allows to insert a number defined via #define into a string */
#define TONUM(Num) #Num
#define NUMTOSTR(S1, Num, S2) S1##TONUM(Num)##S2

/*                     "123456789ABC|F1____F2____F3____F4____F5____F6____F7____F8____F9____F10___F11___F12___" */
HEM_BYTE FnNoShift[] = "010010000000|      Copy              Goto                    Next  Prev              ";

HEM_FNKEYS FnKeys = {FnNoShift, (PHEM_BYTE)"", (PHEM_BYTE)"", (PHEM_BYTE)""};

/* external definitions */
extern bool ProcessVersionInfo(PBYTE lpRes, DWORD ResSize);

//
// Show an error message
//
int ShowError(PCHAR Message)
{
	HiewGate_Message((PHEM_BYTE)"Version Info Error", (PHEM_BYTE)Message);
	return HEM_OK;
}

//
// Unmap file
//
void UnmapFile()
{
	if(lpMapping) UnmapViewOfFile(lpMapping);
	if(hMapping) CloseHandle(hMapping);
	if(hFile) CloseHandle(hFile);
	lpMapping = 0;
	hMapping = 0;
	hFile = 0;
}

//
// Map file
//
bool MapFile(PHEM_BYTE FileName)
{
	hFile = CreateFile((PCHAR)FileName, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	if(hFile != INVALID_HANDLE_VALUE)
	{
		if(hMapping = CreateFileMapping(hFile, 0, PAGE_READONLY, 0, 0, 0))
			if(lpMapping = (LPBYTE)MapViewOfFile(hMapping, FILE_MAP_READ, 0, 0, 0))
				return true;
		UnmapFile();
	}
	else hFile = 0;
	ShowError("Could not access the file");
	return false;
}

//
// Find all VERSIONINFO resources
//
WORD ViSignature[] = {'V', 'S', '_', 'V', 'E', 'R', 'S', 'I', 'O', 'N', '_', 'I', 'N', 'F', 'O', 0, 0, LOWORD(VS_FFI_SIGNATURE), HIWORD(VS_FFI_SIGNATURE)};

char TooManyVIs[] = NUMTOSTR("There are more than ", MAX_VERSIONINFOS, " VERSIONINFO resources");

bool FindAllVIsOld()
{
	TotalVIs = 0;
	int End = FileLength - sizeof(ViSignature);

	/* there are three word-wide fields before the VS_VERSION_INFO identifier so i is initialized to 6 */
	/* making i and End to be int automatically solves the problem when RsrcSize is smaller than sizeof(ViSignature) */
	for(int i = 6; i < End; i++)
	{
		if(memcmp(&lpMapping[i], ViSignature, sizeof(ViSignature)) == 0)
		{
			ViArray[TotalVIs] = i - 6;
			TotalVIs++;
			if(TotalVIs >= MAX_VERSIONINFOS)
			{
				ShowError(TooManyVIs);
				break;
			}
			// i += sizeof(ViSignature); /* uncomment if accurate search is not necessary */
		}
	}
	if(TotalVIs == 0) ShowError("Current file does not contain any VS_VERSIONINFO resources");
	return TotalVIs != 0;
}

DWORD IsSignature(LPBYTE Addr)
{
	if(memcmp(Addr, ViSignature, sizeof(ViSignature)) == 0)
	{
		ViArray[TotalVIs] = Addr - lpMapping - 6;
		TotalVIs++;
		if(TotalVIs >= MAX_VERSIONINFOS)
		{
			ShowError(TooManyVIs);
			return 1;
		}
	}
	return 0;
}

bool FindAllVIs()
{
// __asm int 1;
	TotalVIs = 0;
	if(FileLength < (sizeof(ViSignature) + 6)) return false;

	LPBYTE BlockAddr = lpMapping + 6;
	DWORD BytesRest = FileLength - sizeof(ViSignature) - 6;
	
	HiewGate_MessageWaitOpen((PHEM_BYTE)"Scanning file. Please wait...");

	while(BytesRest)
	{
		DWORD NumBytes = (BytesRest > BYTES_BETWEEN_BREAK) ? BYTES_BETWEEN_BREAK : BytesRest;
		BytesRest -= NumBytes;
		__asm
		{
			mov		edi,BlockAddr
repeat:		mov		ecx,NumBytes
			mov		al,'V'

rescas:		repnz	scasb
			jnz		blockdone
			cmp		byte ptr [edi],0
			jz		fullcheck
			test	ecx,ecx			// repnz scasb does not change the ZF flag if ecx is 0
			jnz		rescas
			jmp		blockdone

fullcheck:	lea		eax,[edi-1]
			mov		NumBytes,ecx
			push	eax
			call	IsSignature
			test	eax,eax
			jz		repeat
			mov		BytesRest,0
		}

blockdone:
		BlockAddr += BYTES_BETWEEN_BREAK;

		/* check for user abort request */
		if(HiewGate_IsKeyBreak() == HEM_KEYBREAK)
		{
			HiewGate_MessageWaitClose();
			HiewGate_Message(0, (PHEM_BYTE)"Aborted by hacker");
			return false;
		}
	}
	
	HiewGate_MessageWaitClose();
	if(TotalVIs == 0) ShowError("Current file does not contain any VS_VERSIONINFO resources");
	return TotalVIs != 0;
}


//
// Let the user select one of found VS_VERSIONINFO resources
// Returns the ordinal number of the selected resource or -1
//
int SelectVI(int CurrentVI)
{
	if(TotalVIs)
	{
		if(TotalVIs == 1) return 0;
		StrArrayFree();
		for(int i = 0; i < TotalVIs; i++)
		{
			char Buffer[512];
			wsprintf(Buffer, "VERSIONINFO resource @ 0x%08X", ViArray[i]);
			StrArrayAdd(Buffer);
		}
		
		int Menu = HiewGate_Menu((PHEM_BYTE)" Select VERSIONINFO resource ", StrArrayGetArray(), StrArrayGetCount(), StrArrayGetWidth(), CurrentVI + 1, 0, 0);
		StrArrayFree();
		if(Menu != HEM_INPUT_ESC && 1 <= Menu && Menu <= TotalVIs) return Menu - 1;
	}
	return -1;
}

//
// Copy decompiled VS_VERSIONINFO to clipboard
//
void CopyToClipboard()
{
	int NumLines = StrArrayGetCount();
	if(NumLines)
	{
		PCHAR *Lines = (PCHAR *)StrArrayGetArray();
		DWORD BlockSize = 0;
		for(int i = 0; i < NumLines; i++) BlockSize += strlen(Lines[i]);
		BlockSize += (NumLines - 1) * 2 + 1; /* CRLF after each line plus terminating zero */
		if(OpenClipboard(0))
		{
			EmptyClipboard();
			HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE, BlockSize);
			if(hMem)
			{
				PCHAR lpMem = (PCHAR)GlobalLock(hMem);
				if(lpMem)
				{
					for(int i = 0; i < NumLines; i++)
					{
						if(i)
						{
							lpMem[0] = 0x0d;
							lpMem[1] = 0x0a;
							lpMem += 2;
						}
						strcpy(lpMem, Lines[i]);
						lpMem += strlen(Lines[i]);
					}
					lpMem[0] = 0;

					GlobalUnlock(hMem);
					SetClipboardData(CF_TEXT, hMem);	/* we must not free the memory after calling SetClipboardData */
				}
				else
				{
					ShowError("GlobalLock() failed");
					GlobalFree(hMem);
				}

			}
			else ShowError("Not enough memory");
			CloseClipboard();
		}
		else ShowError("OpenClipboard() failed");
	}
}

//
// Enable/disable some fucntional keys according to the current situation
//
void SetFnKeys(int CurrentVI)
{
	/* disable both F9 and F10 */
	char F9 = '0';
	char F10 = '0';

	/* check whether F9/F10 may be safely enabled */
	if(TotalVIs > 1)
	{
		if(CurrentVI > 0) F10 = '1';
		if(CurrentVI < (TotalVIs - 1)) F9 = '1';
	}

	FnNoShift[8] =  F9;
	FnNoShift[9] = F10;

	/* enable F6 if there are more than one VERSIONINFO resource */
	FnNoShift[5] = (TotalVIs > 1) ? '1' : '0';
}

//
// Get current file offset
//
DWORD GetCurrentFileOffset()
{
	HIEWGATE_GETDATA HiewData;
	if(HiewGate_GetData(&HiewData) == HEM_OK)
	{
		if(HiewData.offsetCurrent <= 0xffffffff) return (DWORD)HiewData.offsetCurrent;
		else return 0xffffffff;
	}
	else return 0;
}

//
// Convert current file offset to the index of the nearest VI resource in case when there are multiple VIs
//
int FileOffsetToVI()
{
	DWORD FileOffset = GetCurrentFileOffset();
	if(TotalVIs > 1)
	{
		for(int i = TotalVIs - 1; i > 0; i--)
			if(FileOffset >= ViArray[i])
				return i;
	}
	return 0;
}

//
// Plugin entry point
//
int HEM_API Hem_EntryPoint(HEMCALL_TAG *hemCall)
{
	if(hemCall->cbSize < sizeof(HEMCALL_TAG)) return HEM_ERROR;

// __asm int 1;
	
	HIEWGATE_GETDATA HiewData;
	if(HiewGate_GetData(&HiewData) != HEM_OK) return HEM_ERROR;

	if(HiewData.filelength > 0xffffffff)
	{
		ShowError("File is too big");
		return HEM_OK;
	}
	FileLength = (DWORD)HiewData.filelength;
	
	if(MapFile(HiewData.filename))
	{
		if(FindAllVIs())
		{
			int CurrentVI = FileOffsetToVI();

reselect:	CurrentVI = SelectVI(CurrentVI);
			if(CurrentVI != -1)
			{
newvi:			DWORD ResOffset = ViArray[CurrentVI];
				LPBYTE lpRes = ResOffset + lpMapping;
				DWORD ResSize = (lpRes[1] << 8) + lpRes[0];	/* the first word of VS_VERSIONINFO is the length of the compiled resource */
			
				/* check that we do not cross the end of the buffer */
				DWORD BytesRest = FileLength - ResOffset;
				if(ResSize > BytesRest) ResSize = BytesRest;
			
				StrArrayFree();
				bool ValidVI = ProcessVersionInfo(lpRes, ResSize);
				
				/* show at least something so that we can at least press F5 and go to the VI structure even if it is totally invalid */
				if(StrArrayGetCount() == 0) StrArrayAdd("Invalid VERSION_INFO structure");
				
				HEM_BYTE Title[512];
				HEM_UINT FnKey = 0;
			
				int Width = StrArrayGetWidth();

				int TitleLen = wsprintf((PCHAR)Title, " %sVersion Info Resource @ 0x%08X ", ValidVI ? "" : "Invalid ", ResOffset);
				if(TitleLen > Width) Width = TitleLen;
				
				SetFnKeys(CurrentVI);
				HiewGate_Window(Title, StrArrayGetArray(), StrArrayGetCount(), Width, &FnKeys, &FnKey);
			
				/* F6/F9/F10 keys get disabled in SetFnKeys() according to the current situation so we may skip some (or all) important checks (like crossing over the boundaries) here */
				switch(FnKey)
				{
					case HEM_FNKEY_F5:
						hemCall->returnActionFlag |= HEM_RETURN_SETOFFSET;
						hemCall->returnOffset = ResOffset;
						break;

					case HEM_FNKEY_F2:
						CopyToClipboard();
						break;

					case HEM_FNKEY_F9:
						CurrentVI++;
						goto newvi;

					case HEM_FNKEY_F10:
						CurrentVI--;
						goto newvi;

					case HEM_FNKEY_F6:
						goto reselect;
				}
			}
		}
	}
	StrArrayFree();
	UnmapFile();
	return HEM_OK;
}

//
// Plugin loading stuff
//
HEMINFO_TAG hemVerInfo = {sizeof(HEMINFO_TAG),
                          sizeof(int),
						  0,						// reserved
						  HEM_SDK_VERSION_MAJOR, HEM_SDK_VERSION_MINOR,
						  VERINFO_VERSION_MAJOR, VERINFO_VERSION_MINOR,
						  (HEM_FLAG_FILEMASK & ~HEM_FLAG_DISK) | HEM_FLAG_MODEMASK,
						  0,						// reserved
						  Hem_EntryPoint,
						  NULL,						// no Hem_Unload()
						  NULL,						// no Hem2HemGate()
						  0,						// reserved
						  0,						// reserved
						  0,						// reserved
						  0,						// reserved
						  "Version Info Ex",
						  "VERSIONINFO Resource Viewer (Extended Version)",
						  "*****************************************",
						  "* FSC (c) 2005 http://www.fsc-soft.com/ *",
						  "*****************************************"};



int HEM_EXPORT Hem_Load(HIEWINFO_TAG *hiewInfo)
{
	/* check that SDK version is OK */
	DWORD SDKVersion = (hiewInfo->sdkVerMajor << 8) | hiewInfo->sdkVerMinor;
	DWORD RequiredSDKVersion = (HEM_SDK_VERSION_MAJOR << 8) | HEM_SDK_VERSION_MINOR;
	if(SDKVersion < RequiredSDKVersion) return HEM_ERR_SDKVER_INCOMPATIBLE;

	HiewGate_Set(hiewInfo);
	hiewInfo->hemInfo = &hemVerInfo;
	return HEM_OK;
}

//
// DLL main entry point
//
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	switch(fdwReason)
	{
		case DLL_PROCESS_ATTACH:
			StrArrayInitialize();
			DisableThreadLibraryCalls(hinstDLL);
			break;

		case DLL_PROCESS_DETACH:
			StrArrayFree();
			break;
	}
	return TRUE;
}
