//
//  High-level Hiew Gate interface
//
#include "hem.h"
#include <string.h>

////////////////////////////////////////////////////////////
// Address of the HiewGate function, received by Hem_Load()

int (HEM_API *HiewGate)( void *tag ) = NULL;

////////////////////////////////////////////////////////////
// Hem handle, received by Hem_Load() 

int hemHandle=0;

////////////////////////////////////////////////////////////
// Set HiewGate and hemHandle, call from Hem_Load() 

int
 HiewGate_Set( HIEWINFO_TAG *hiewInfo )
{
 HiewGate=hiewInfo->HiewGate; 
 hemHandle=hiewInfo->hemHandle;
 return( HEM_OK );
}

////////////////////////////////////////////////////////////
// Call of the HiewGate

int
 HiewGateHighLevel( void *tag )
{
 ((HIEWGATE_NULL*)tag)->hemHandle=hemHandle;
 return( HiewGate? HiewGate( tag ) : HEM_ERR_NOADDRESS_HIEWGATE );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_ERROR          
//    HEM_OK

int
 HiewGate_Null( void )
{
 HIEWGATE_NULL
  tag={ sizeof( HIEWGATE_NULL ), HIEWGATE_ID_NULL };

 return( HiewGateHighLevel( &tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_ERR_HIEWDATA_SIZE_MISMATCH          
//    HEM_OK

int
 HiewGate_GetData( HIEWGATE_GETDATA *tag )
{
 tag->cbSize=sizeof( HIEWGATE_GETDATA );
 tag->callId=HIEWGATE_ID_GETDATA;
 return( HiewGateHighLevel( tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_HEM_ERR_HEM2HEMGATE_IS_NULL
//    HEM_ERR_HEM_NOTFOUND          
//    HEM_OK

int
 HiewGate_GetHem2HemGate( HIEWGATE_GETHEM2HEMGATE *tag, HEM_BYTE *shortName )
{
 tag->cbSize=sizeof( HIEWGATE_GETHEM2HEMGATE );
 tag->callId=HIEWGATE_ID_GETHEM2HEMGATE;
 strncpy( tag->shortName, shortName, HEM_SHORTNAME_SIZE-1 );
 return( HiewGateHighLevel( tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    NULL on error
//    pointer

HEM_BYTE
 *HiewGate_GetMemory( HEM_UINT bytes )
{
 HIEWGATE_GETMEMORY
  tag={ sizeof( HIEWGATE_GETMEMORY ), HIEWGATE_ID_GETMEMORY };

 tag.bytes=bytes;
 tag.retPmem=NULL;
 HiewGateHighLevel( &tag );
 return( tag.retPmem );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_OK

int
 HiewGate_FreeMemory( HEM_BYTE *pMem )
{
 HIEWGATE_FREEMEMORY
  tag={ sizeof( HIEWGATE_FREEMEMORY ), HIEWGATE_ID_FREEMEMORY };

 tag.pMem=pMem;
 return( HiewGateHighLevel( &tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_ERR_READONLYFILE
//    HEM_OK

int
 HiewGate_FileOpenForWrite()
{
 HIEWGATE_FILEOPENFORWRITE
  tag={ sizeof( HIEWGATE_FILEOPENFORWRITE ), HIEWGATE_ID_FILEOPENFORWRITE };

 return( HiewGateHighLevel( &tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_ERR_POINTER_IS_NULL
//    HEM_ERROR
//    number of bytes read

int
 HiewGate_FileRead( HEM_QWORD offset, HEM_UINT bytes, HEM_BYTE *buffer )
{
 HIEWGATE_FILEREAD
  tag={ sizeof( HIEWGATE_FILEREAD ), HIEWGATE_ID_FILEREAD };

 tag.offset=offset;
 tag.bytes=bytes;
 tag.buffer=buffer;
 return( HiewGateHighLevel( &tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_ERR_POINTER_IS_NULL
//    HEM_ERROR
//    number of bytes written

int
 HiewGate_FileWrite( HEM_QWORD offset, HEM_UINT bytes, HEM_BYTE *buffer )
{
 HIEWGATE_FILEWRITE
  tag={ sizeof( HIEWGATE_FILEWRITE ), HIEWGATE_ID_FILEWRITE };

 tag.offset=offset;
 tag.bytes=bytes;
 tag.buffer=buffer;
 return( HiewGateHighLevel( &tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    key pressed in upper case

int
 HiewGate_Message( HEM_BYTE *title, HEM_BYTE *msg )
{
 HIEWGATE_MESSAGE
  tag={ sizeof( HIEWGATE_MESSAGE ), HIEWGATE_ID_MESSAGE };

 tag.title=title;
 tag.msg=msg;
 return( HiewGateHighLevel( &tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_ERR_FNKEYS_INVALID          
//    HEM_INPUT_ESC
//    HEM_OK
//    pressed FnKey stored into returnFnKey

int
 HiewGate_Window( HEM_BYTE *title, HEM_BYTE **lines, int linesCount, int width, HEM_FNKEYS *fnKeys, HEM_UINT *returnFnKey )
{
 int
  rc;
 HIEWGATE_WINDOW
  tag={ sizeof( HIEWGATE_WINDOW ), HIEWGATE_ID_WINDOW };

 tag.title=title;
 tag.lines=lines;
 tag.linesCount=linesCount;
 tag.width=width;
 if( fnKeys ){
   tag.fnKeys.main=fnKeys->main;
   tag.fnKeys.alt=fnKeys->alt;
   tag.fnKeys.ctrl=fnKeys->ctrl;
   tag.fnKeys.shift=fnKeys->shift;
 }else{
   tag.fnKeys.main=NULL;
   tag.fnKeys.alt=NULL;
   tag.fnKeys.ctrl=NULL;
   tag.fnKeys.shift=NULL;
 }
 rc=HiewGateHighLevel( &tag );
 if( returnFnKey )
   *returnFnKey=tag.returnFnKey;
 return( rc );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_ERR_FNKEYS_INVALID          
//    HIEWGATE_INPUT_ESC
//    selected item index, starting from 1
//    pressed FnKey stored into returnFnKey

int
 HiewGate_Menu( HEM_BYTE *title, HEM_BYTE **lines, int linesCount, int width, int startItem, HEM_FNKEYS *fnKeys, HEM_UINT *returnFnKey )
{
 int
  rc;
 HIEWGATE_MENU
  tag={ sizeof( HIEWGATE_MENU ), HIEWGATE_ID_MENU };

 tag.title=title;
 tag.lines=lines;
 tag.linesCount=linesCount;
 tag.width=width;
 tag.startItem=startItem;   // [ 1 ... linesCount ]
 if( fnKeys ){
   tag.fnKeys.main=fnKeys->main;
   tag.fnKeys.alt=fnKeys->alt;
   tag.fnKeys.ctrl=fnKeys->ctrl;
   tag.fnKeys.shift=fnKeys->shift;
 }else{
   tag.fnKeys.main=NULL;
   tag.fnKeys.alt=NULL;
   tag.fnKeys.ctrl=NULL;
   tag.fnKeys.shift=NULL;
 }
 rc=HiewGateHighLevel( &tag );
 if( returnFnKey )
   *returnFnKey=tag.returnFnKey;
 return( rc );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_ERROR          
//    HIEWGATE_INPUT_ESC
//    HIEWGATE_INPUT_CR

int
 HiewGate_GetString( HEM_BYTE *title, HEM_BYTE *string, int stringLen )
{
 HIEWGATE_GETSTRING
  tag={ sizeof( HIEWGATE_GETSTRING ), HIEWGATE_ID_GETSTRING };

 tag.title=title;
 tag.string=string;
 tag.stringLen=stringLen;
 return( HiewGateHighLevel( &tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_OK

int
 HiewGate_MessageWaitOpen( HEM_BYTE *msg )
{
 HIEWGATE_MESSAGEWAITOPEN
  tag={ sizeof( HIEWGATE_MESSAGEWAITOPEN ), HIEWGATE_ID_MESSAGEWAITOPEN };

 tag.msg=msg;
 return( HiewGateHighLevel( &tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_OK

int
 HiewGate_MessageWaitClose()
{
 HIEWGATE_MESSAGEWAITCLOSE
  tag={ sizeof( HIEWGATE_MESSAGEWAITCLOSE ), HIEWGATE_ID_MESSAGEWAITCLOSE };

 return( HiewGateHighLevel( &tag ) );
}

////////////////////////////////////////////////////////////
// returns: 
//    HEM_KEYBREAK, if ESC was pressed

int
 HiewGate_IsKeyBreak()
{
 HIEWGATE_ISKEYBREAK
  tag={ sizeof( HIEWGATE_ISKEYBREAK ), HIEWGATE_ID_ISKEYBREAK };

 return( HiewGateHighLevel( &tag ) );
}

/// Module End /////////////////////////////////////////////

